from typing import Tuple

import pandas as pd
import numpy as np
import matplotlib
from matplotlib.lines import Line2D
import matplotlib.pyplot as plt
import igraph as ig

from decoupler._docs import docs
from decoupler._Plotter import Plotter


def _src_idxs(
    score: pd.DataFrame,
    sources: int | list | str,
    by_abs: bool,
) -> np.ndarray:
    assert isinstance(sources, (int, list, str)), \
    'sources must be int, list or str'
    if isinstance(sources, int):
        if by_abs:
            s_idx = np.argsort(-abs(score.values[0]))[:sources]
        else:
            s_idx = np.argsort(-score.values[0])[:sources]
    elif isinstance(sources, list):
        s_idx = np.isin(score.columns.astype(str), sources)
    else:
        s_idx = np.isin(score.columns.astype(str), [sources])
    return s_idx


def _trg_idxs(
    data: pd.DataFrame,
    net: pd.DataFrame,
    targets: int | list | str,
    by_abs: bool,
) -> np.ndarray:
    assert isinstance(targets, (int, list, str)), \
    'targets must be int, list or str'
    if isinstance(targets, int):
        net['prod'] = [data.iloc[0][t] * w if t in data.columns else 0 for t, w in zip(net['target'], net['weight'])]
        if by_abs:
            net['prod'] = abs(net['prod'])
        t_idx = (
            net
            .sort_values(['source', 'prod'], ascending=[True, False])
            .groupby(['source'], observed=True)
            .head(targets)
            .index
            .values
        )
    elif isinstance(targets, list):
        t_idx = np.isin(net['target'].astype(str), targets)
    else:
        t_idx = np.isin(net['target'].astype(str), [targets])
    return t_idx

    
def _filter(
    data: pd.DataFrame,
    score: pd.DataFrame,
    net: pd.DataFrame,
    sources: int,
    targets: int,
    by_abs: bool,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    assert isinstance(data, pd.DataFrame), 'data must be pd.DataFrame'
    assert isinstance(score, pd.DataFrame), 'score must be pd.DataFrame'
    assert np.all(data.index == score.index) and (data.index.size == 1), \
    'data and score need to have the same row index.'
    assert isinstance(by_abs, bool), 'by_abs must be bool'
    # Select top sources
    s_idx = _src_idxs(score=score, sources=sources, by_abs=by_abs)
    # Filter
    score = score.iloc[:, s_idx]
    net = net.loc[np.isin(net['source'].astype(str), score.columns.astype(str)), :].copy()
    if 'weight' not in net.columns:
        net['weight'] = 1.
    # Select top targets
    t_idx = _trg_idxs(data=data, net=net, targets=targets, by_abs=by_abs)
    # Filter
    net = net.loc[t_idx]
    # Filter unmatched features
    data = data.loc[:, np.isin(data.columns.astype(str), net['target'].astype(str))]
    net = net.loc[np.isin(net['target'].astype(str), data.columns.astype(str)), :]
    return data, score, net


def _norm(
    x: np.ndarray,
    vcenter: bool,
) -> matplotlib.colors.Normalize:
    assert isinstance(vcenter, bool), 'vcenter must be bool'
    if vcenter:
        vmax = np.max(np.abs(x))
        norm = matplotlib.colors.Normalize(vmin=-vmax, vmax=vmax)
    else:
        vmax = np.max(x)
        vmin = np.min(x)
        norm = matplotlib.colors.Normalize(vmin=vmin, vmax=vmax)
    return norm


def _dict_types(
    data: pd.DataFrame,
    score: pd.DataFrame,
) -> Tuple[dict, np.ndarray]:
    vs = np.unique(np.hstack([data.columns, score.columns]))
    v_dict = {k: i for i, k in enumerate(vs)}
    types = (~np.isin(vs, score.columns)) * 1
    return v_dict, types


def _net_2_elist(
    net: pd.DataFrame,
    v_dict: dict,
) -> list:
    edges = []
    for i in net.index:
        source, target = net.loc[i, 'source'], net.loc[i, 'target']
        edge = [v_dict[source], v_dict[target]]
        edges.append(edge)
    return edges


def _net_2_g(
    data: pd.DataFrame,
    score: pd.DataFrame,
    net: pd.DataFrame,
) -> ig.Graph:
    # Unify network
    v_dict, types = _dict_types(data=data, score=score)
    # Transform net to edges
    edges = _net_2_elist(net=net, v_dict=v_dict)
    # Create graph
    g = ig.Graph(
        edges=edges,
        directed=True,
    )
    # Update attributes
    g.es['weight'] = net['weight'].values
    g.vs['type'] = types
    g.vs['label'] = list(v_dict.keys())
    g.vs['shape'] = np.where(types, 'circle', 'square')
    return g


def _gcolors(
    g: ig.Graph,
    data: pd.DataFrame,
    score: pd.DataFrame,
    s_norm: matplotlib.colors.Normalize,
    t_norm: matplotlib.colors.Normalize,
    s_cmap: str,
    t_cmap: str,
) -> bool:
    cmaps = matplotlib.colormaps.keys()
    if (s_cmap in cmaps) and (t_cmap in cmaps):
        s_cmap = matplotlib.colormaps.get_cmap(s_cmap)
        t_cmap = matplotlib.colormaps.get_cmap(t_cmap)
        color = []
        for i, k in enumerate(g.vs['label']):
            if g.vs['type'][i]:
                color.append(t_cmap(t_norm(data[k].values[0])))
            else:
                color.append(s_cmap(s_norm(score[k].values[0])))
        is_cmap = True
    else:
        color = [s_cmap if typ == 0. else t_cmap for typ in g.vs['type']]
        is_cmap = False
    g.vs['color'] = color
    return is_cmap


@docs.dedent
def network(
    net,
    data: pd.DataFrame = None,
    score: pd.DataFrame = None,
    sources: int | list | str = 5,
    targets: int | list | str = 10,
    by_abs=True,
    size_node=5,
    size_label=2.5,
    s_cmap='RdBu_r',
    t_cmap='viridis',
    vcenter=False,
    c_pos_w='darkgreen',
    c_neg_w='darkred',
    s_label='Enrichment\nscore',
    t_label='Gene\nexpression',
    layout='kk',
    **kwargs
):
    """
    Plot results of enrichment analysis as network.

    Parameters
    ----------
    %(net)s
    data
        Input of enrichment analysis, needs to be a one row dataframe with targets as features. Used to filter net.
    score
        Ouput of enrichment analysis, needs to be a one row dataframe with sources as features. Used to filter net.
    sources
        Number of top sources to plot or list of source names.
    targets
        Number of top targets to plot or list of target names.
    by_abs
        Whether to consider the absolute value when sorting for ``n_sources`` and ``n_targets``.
    size_node
        Size of the nodes in the plot.
    size_label
        Size of the labels in the plot.
    s_cmap
        Color or colormap to use to color sources.
    t_cmap
        Color or colormap to use to color targets.
    vcenter
        Whether to center colors around 0.
    c_pos_w
        Color for edges with positive weights. If no weights are available, they are set to positive by default.
    c_neg_w
        Color for edges with negative weights.
    s_label
        Label to place in the source colorbar.
    t_label
        Label to place in the target colorbar.
    layout
        Layout to use to order the nodes. Check ``igraph`` documentation for more options.
    %(plot)s
    """
    assert isinstance(net, pd.DataFrame), 'net must be pd.DataFrame'
    assert (data is None) == (score is None), 'data and score must either both be None'
    if data is None:
        srcs = net['source'].unique().astype('U')
        score = pd.DataFrame(np.ones((1, srcs.size)), index=['0'], columns=srcs)
        trgs = net['target'].unique().astype('U')
        data = pd.DataFrame(np.ones((1, trgs.size)), index=['0'], columns=trgs)
        t_cmap = 'white'
    # Filter
    fdata, fscore, fnet = _filter(
        data=data,
        score=score,
        net=net,
        sources=sources,
        targets=targets,
        by_abs=by_abs,
    )
    # Color norms
    s_norm = _norm(x=fscore, vcenter=vcenter)
    t_norm = _norm(x=fdata, vcenter=vcenter)
    # Make graph
    g = _net_2_g(data=fdata, score=fscore, net=fnet)
    g.es['color'] = [c_pos_w if w > 0 else c_neg_w for w in g.es['weight']]
    is_cmap = _gcolors(
        g=g,
        data=data,
        score=score,
        s_norm=s_norm,
        t_norm=t_norm,
        s_cmap=s_cmap,
        t_cmap=t_cmap,
    )
    # Instance
    kwargs['ax'] = None
    bp = Plotter(**kwargs)
    bp.fig.delaxes(bp.ax)
    # Plot
    gs = matplotlib.gridspec.GridSpec(6, 3)
    ax1 = bp.fig.add_subplot(gs[:-1, :])
    ax2 = bp.fig.add_subplot(gs[-1, 0])
    ax3 = bp.fig.add_subplot(gs[-1, 1])
    ax4 = bp.fig.add_subplot(gs[-1, -1])
    ig.plot(
        g,
        target=ax1,
        layout=layout,
        vertex_size=(size_node * bp.dpi) / (bp.figsize[0] * bp.figsize[0]),
        vertex_size_label=(size_label * bp.dpi) / (bp.figsize[0] * bp.figsize[0]),
        bbox_inches='tight',
    )
    if is_cmap:
        sm = matplotlib.cm.ScalarMappable(norm=s_norm, cmap=s_cmap)
        bp.fig.colorbar(sm, cax=ax2, orientation="horizontal", label=s_label)
        sm = matplotlib.cm.ScalarMappable(norm=t_norm, cmap=t_cmap)
        bp.fig.colorbar(sm, cax=ax4, orientation="horizontal", label=t_label)
    else:
        ax2.axis("off")
        ax4.axis("off")
    # Add legend
    square = Line2D([0], [0], marker='s', color='w', label='Source', markerfacecolor='white',
                    markeredgecolor='black', markersize=10)
    circle = Line2D([0], [0], marker='o', color='w', label='Target', markerfacecolor='white',
                    markeredgecolor='black', markersize=10)
    line1 = Line2D((0, 0), (1, 0), color=c_pos_w, lw=2, marker='>',)
    line2 = Line2D((0, 0), (1, 0), color=c_neg_w, lw=2, marker='>',)
    handles = [square, circle, line1, line2]
    labels = ['Source', 'Target', 'Positive', 'Negative']
    legend = ax3.legend(
        handles=[square, circle, line1, line2],
        labels=labels,
        frameon=False,
        loc='center', bbox_to_anchor=(0.5, 0.5), bbox_transform=ax3.transAxes
    )
    ax3.axis('off')
    return bp._return()



import json

import requests
import pandas as pd

from decoupler._docs import docs
from decoupler._log import _log
from decoupler._download import URL_DBS, _download
from decoupler.op._translate import translate
from decoupler.op._dtype import _infer_dtypes


def show_resources(
) -> pd.DataFrame:
    """
    Shows available resources in Omnipath :cite:p:`omnipath`.
    For more information visit the official
    [website](https://omnipathdb.org/).

    Returns
    -------
    List of available resources to query with `decoupler.op.resource`.
    """
    ann = pd.read_csv('https://omnipathdb.org/queries/annotations', sep='\t')
    ann = ann.set_index('argument').loc['databases'].str.split(';')['values']
    url = 'https://omnipathdb.org/resources'
    response = requests.get(url)
    lcs = response.json()
    df = pd.DataFrame(ann, columns=['name'])
    df['license'] = [lcs[a]['license']['purpose'] if a in lcs else None for a in ann]
    return df


@docs.dedent
def resource(
    name: str,
    organism: str = 'human',
    license: str = 'academic',
    verbose: bool = False,
):
    """
    Wrapper to access resources inside Omnipath :cite:p:`omnipath`.

    This wrapper allows to easly query different prior knowledge resources. To
    check available resources run ``decoupler.op.show_resources()``. For more
    information visit the official [website](https://omnipathdb.org/).

    Parameters
    ----------
    name:
        Name of the resource to query.
    %(organism)s
    %(license)s
    %(verbose)s
    kwargs
        Passed to ``decoupler.op.translate``.

    Returns
    -------
    Network in long format.
    """
    # Validate
    assert isinstance(name, str), 'name must be str'
    names = set(show_resources()['name'])
    assert name in names, f'name must be one of these: {names}'
    assert isinstance(organism, str), 'organism must be str'
    assert isinstance(license, str) and license in ['academic', 'commercial', 'nonprofit'], \
    'license must be academic, commercial or nonprofit'
    assert isinstance(verbose, bool), 'verbose must be bool'
    m = f'Accessing {name} with {license} license'
    _log(m, level='info', verbose=verbose)
    # Download
    url = URL_DBS + f'{name}&license={license}'
    df = _download(url, sep='\t', verbose=verbose)
    # Process
    labels = df['label'].unique()
    for label in labels:
        if label in df.columns:
            df.loc[df['label'] == label, 'label'] = f'_{label}'
    df = df[['genesymbol', 'label', 'value', 'record_id']]
    df = df.pivot(index=["genesymbol", "record_id"], columns="label", values="value").reset_index()
    df.index.name = ''
    df.columns.name = ''
    cols_to_remove = ['record_id', 'entity_type', '_entity_type']
    df = df.drop(columns=[c for c in cols_to_remove if c in df.columns])
    df = _infer_dtypes(df)
    if organism != 'human':
        df = translate(df, columns='genesymbol', target_organism=organism, verbose=verbose)
    return df



import numpy as np
import pandas as pd
import pytest

import decoupler as dc


def test_show_organisms():
    lst = dc.op.show_organisms()
    assert isinstance(lst, list)
    assert len(lst) > 0
    assert {'mouse', 'rat'}.issubset(lst)


@pytest.mark.parametrize(
    'lst,my_dict,one_to_many',
    [
        [['a', 'b', 'c', 'd'], dict(a=['B', 'C'], b=['A', 'C'], c=['A', 'B'], d='D'), 1],
        [['a', 'b', 'c', 'd'], dict(c=['A', 'B']), 1],
        [['a', 'b', 'c', 'd'], dict(a=['B', 'C'], b=['A', 'C'], c=['A', 'B'], d='D'), 10],
    ]
)
def test_replace_subunits(
    lst,
    my_dict,
    one_to_many,
):  
    res = dc.op._translate._replace_subunits(
        lst=lst, my_dict=my_dict, one_to_many=one_to_many
    )
    assert isinstance(res, list)
    assert len(res) == len(lst)
    for k in my_dict:
        idx = lst.index(k)
        if k in my_dict:
            if len(my_dict[k]) > one_to_many:
                assert np.isnan(res[idx])
            else:
                assert isinstance(res[idx], list)
        else:
            assert np.isnan(res[idx])


@pytest.mark.parametrize('target_organism', ['mouse', 'anole_lizard', 'fruitfly'])
def test_translate(
    target_organism,
):
    net = dc.op.collectri()
    t_net = dc.op.translate(net=net, columns='target', target_organism='mouse')
    cols = {'source', 'target', 'weight'}
    assert isinstance(t_net, pd.DataFrame)
    assert cols.issubset(t_net.columns)
    assert net.shape[0] != t_net.shape[0]









import pandas as pd
import numpy as np
import pytest

import decoupler as dc


@pytest.mark.parametrize('remove_complexes', [True, False])
def test_collectri(
    remove_complexes,
):
    ct = dc.op.collectri(remove_complexes=remove_complexes)
    assert isinstance(ct, pd.DataFrame)
    cols = {'source', 'target', 'weight', 'resources', 'references', 'sign_decision'}
    assert cols.issubset(ct.columns)
    assert pd.api.types.is_numeric_dtype(ct['weight'])
    msk = np.isin(['AP1', 'NFKB'], ct['source']).all()
    if remove_complexes:
        assert not msk
    else:
        assert msk
    assert not ct.duplicated(['source', 'target']).any()



import numpy as np
from matplotlib.figure import Figure
import seaborn as sns
from anndata import AnnData

from decoupler._docs import docs
from decoupler._Plotter import Plotter
from decoupler.pp.data import extract
from decoupler.pp.anndata import _min_sample_size, _ssize_tcount


@docs.dedent
def filter_by_expr(
    adata: AnnData,
    group: str | None = None,
    lib_size: float | None = None,
    min_count: int = 10,
    min_total_count: int = 15,
    large_n: int = 10,
    min_prop: float = 0.7,
    cmap: str = 'viridis',
    **kwargs,
) -> None | Figure:
    """
    Plot to help determining the thresholds of the ``decoupler.pp.filter_by_expr`` function.

    Parameters
    ----------
    %(adata)s
    %(cmap)s
    %(group)s
    %(lib_size)s
    %(min_count)s
    %(min_total_count)s
    %(large_n)s
    %(min_prop_expr)s
    %(plot)s
    """
    assert isinstance(adata, AnnData), 'adata must be AnnData'
    # Extract inputs
    X, _, _ = extract(adata, empty=False)
    obs = adata.obs
    # Minimum sample size cutoff
    min_sample_size = _min_sample_size(
        obs=obs,
        group=group,
        large_n=large_n,
        min_prop=min_prop,
    )
    # Compute sample size and total count
    sample_size, total_count = _ssize_tcount(
        X=X,
        lib_size=lib_size,
        min_count=min_count,
    )
    # Total counts
    total_count[total_count < 1.] = np.nan  # Handle 0s
    # Instance
    bp = Plotter(**kwargs)
    # Plot
    sns.histplot(
        x=np.log10(total_count),
        y=sample_size,
        cmap=cmap,
        cbar=True,
        cbar_kws=dict(shrink=.75, label='Number of genes'),
        discrete=(False, True),
        ax=bp.ax,
    )
    bp.ax.axhline(y=min_sample_size - 0.5, c='gray', ls='--')
    bp.ax.axvline(x=np.log10(min_total_count), c='gray', ls='--')
    bp.ax.set_xlabel(r'$\log_{10}$ total sum of counts')
    bp.ax.set_ylabel('Number of samples')
    return bp._return()



import pandas as pd
import numpy as np
import pytest

import decoupler as dc


def test_show_resources():
    df = dc.op.show_resources()
    assert isinstance(df, pd.DataFrame)
    assert df.shape[0] > 0
    assert {'name', 'license'}.issubset(df.columns)
    assert np.isin(['PROGENy', 'MSigDB'], df['name']).all()


@pytest.mark.parametrize('name', ['Lambert2018', 'PanglaoDB'])
def test_resource(
    name
):
    rs = dc.op.resource(name=name)
    assert isinstance(rs, pd.DataFrame)
    assert 'genesymbol' in rs.columns



import numpy as np
import pandas as pd

from decoupler._docs import docs
from decoupler._log import _log
from decoupler._download import URL_INT, _download
from decoupler.op._translate import translate
from decoupler.op._dtype import _infer_dtypes


@docs.dedent
def hallmark(
    organism: str = 'human',
    license: str = 'academic',
    verbose: bool = False,
) -> pd.DataFrame:
    """
    Hallmark gene sets :cite:p:`msigdb`.

    Hallmark gene sets summarize and represent specific well-defined
    biological states or processes and display coherent expression.

    Parameters
    ----------
    %(organism)s
    %(license)s
    %(verbose)s

    Returns
    -------
    Dataframe in long format containing the hallmark gene sets.
    """
    url = 'https://static.omnipathdb.org/tables/msigdb-hallmark.tsv.gz'
    hm = _download(url, sep='\t', compression='gzip', verbose=verbose)
    hm = hm[['geneset', 'genesymbol']]
    hm['geneset'] = hm['geneset'].str.replace('HALLMARK_', '')
    hm['genesymbol'] = hm['genesymbol'].str.replace('COMPLEX:', '').str.split('_')
    hm = hm.explode('genesymbol')
    hm = _infer_dtypes(hm)
    if organism != 'human':
        hm = translate(hm, columns=['genesymbol'], target_organism=organism, verbose=verbose)
    hm = hm.rename(columns={'geneset': 'source', 'genesymbol': 'target'})
    hm = hm.drop_duplicates(['source', 'target']).reset_index(drop=True)
    return hm



from typing import Tuple

import numpy as np

from decoupler.bm._pp import _validate_bool


def fscore(
    y_true: np.ndarray,
    y_score: np.ndarray,
) -> Tuple[float, float, float]:
    """
    F-beta score
    """
    # Validate
    _validate_bool(y_true=y_true, y_score=y_score)
    assert y_score.dtype == np.bool_, \
    'y_score must be bool numpy.ndarray'
    y_true = y_true.astype(np.bool_)
    # Compute
    tp = np.sum(y_true * y_score)
    fp = np.sum((~y_true) * y_score)
    fn = np.sum(y_true * (~y_score))
    if tp > 0:
        prc = tp / (tp + fp)
        rcl = tp / (tp + fn)
    else:
        prc = 0.
        rcl = 0.
    return prc, rcl

fscore.scores = ['precision', 'recall']



import matplotlib.pyplot as plt
from matplotlib.axes._axes import Axes
from matplotlib.figure import Figure

from decoupler._docs import docs


class Plotter:
    @docs.dedent
    def __init__(
        self,
        ax: Axes | None = None,
        figsize: tuple | None = (4, 3),
        dpi: int = 100,
        return_fig: bool = False,
        save: str | None = None,
    ) -> Figure | None:
        """
        Base class for plotters.

        Parameters
        ----------
        %(plot)s
        """
        # Validate
        assert isinstance(ax, Axes) or ax is None, \
        'ax must be matplotlib.axes._axes.Axes or None'
        assert isinstance(figsize, tuple), \
        'figsize must be tuple'
        assert isinstance(dpi, (int, float)) and dpi > 0, \
        'dpi must be numerical and > 0'
        assert isinstance(return_fig, bool), \
        'return_fig must be bool'
        assert isinstance(save, str) or save is None, \
        'save must be str or None'
        self.ax = ax
        self.figsize = figsize
        self.dpi = dpi
        self.return_fig = return_fig
        self.save = save
        if self.ax is None:
            self.fig, self.ax = plt.subplots(1, 1, figsize=self.figsize, dpi=self.dpi, tight_layout=True)
        else:
            self.fig = self.ax.figure

    def _return(self):
        if self.save is not None:
            self.fig.savefig(self.save, bbox_inches='tight')
        if self.return_fig:
            return self.fig



from typing import Tuple

import numpy as np
import numba as nb
import scipy.stats as sts

from decoupler._docs import docs
from decoupler._log import _log
from decoupler._Method import MethodMeta, Method


@nb.njit(parallel=True, cache=True)
def _tval(X, y, inv, df):
    X = np.ascontiguousarray(X)
    n_samples = y.shape[1]
    n_fsets = X.shape[1]
    coef, sse, _, _ = np.linalg.lstsq(X, y)
    assert len(sse) > 0, 'Could not fit a multivariate linear model. This can happen because there are more sources\n \
    (covariates) than unique targets (samples), or because the network adjacency matrix rank is smaller than the number\n \
    of sources'
    sse = sse / df
    se = np.zeros((n_samples, n_fsets))
    for i in nb.prange(n_samples):
        se[i] = np.sqrt(np.diag(sse[i] * inv))
    t = coef.T / se
    return t.astype(nb.f4)


@docs.dedent
def _func_mlm(
    mat: np.ndarray,
    adj: np.ndarray,
    verbose: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    r"""
    Multivariate Linear Model (MLM) :cite:`decoupler`.

    This approach uses the molecular features from one observation as the population of samples
    and it fits a linear model with with multiple covariates, which are the weights of all feature sets :math:`F`.

    .. math::

        y^i = \beta_0 + \beta_1 x_{1}^{i} + \beta_2 x_{2}^{i} + \cdots + \beta_p x_{p}^{i} + \varepsilon

    Where:

    - :math:`y^i` is the observed feature statistic (e.g. gene expression, :math:`log_{2}FC`, etc.) for feature :math:`i`
    - :math:`x_{p}^{i}` is the weight of feature :math:`i` in feature set :math:`F_p`. For unweighted sets, membership in the set is indicated by 1, and non-membership by 0.
    - :math:`\beta_0` is the intercept
    - :math:`\beta_p` is the slope coefficient for feature set :math:`F_p`
    - :math:`\varepsilon` is the error term for feature :math:`i`

    .. figure:: /_static/images/mlm.png
       :alt: Multivariate Linear Model (MLM) schematic.
       :align: center
       :width: 75%

       Multivariate Linear Model (MLM) scheme.
       In this example, the observed gene expression of :math:`Sample_1` is predicted using
       the interaction weights of two pathways, :math:`P_1` and :math:`P_2`.
       For :math:`P2`, since its target genes that have negative weights are lowly expressed,
       and its positive target genes are highly expressed,
       the relationship between the two variables is positive so the obtained :math:`ES` score is positive.
       Scores can be interpreted as active when positive, repressive when negative, and inconclusive when close to 0.

    The enrichment score :math:`ES` for each :math:`F` is then calculated as the t-value of the slope coefficients.

    .. math::

        ES = t_{\beta_1} = \frac{\hat{\beta}_1}{\mathrm{SE}(\hat{\beta}_1)}

    Where:

    - :math:`t_{\beta_1}` is the t-value of the slope
    - :math:`\mathrm{SE}(\hat{\beta}_1)` is the standard error of the slope

    Next, :math:`p_{value}` are obtained by evaluating the two-sided survival function
    (:math:`sf`) of the Student’s t-distribution.

    .. math::

        p_{value} = 2 \times \mathrm{sf}(|ES|, \text{df})

    %(params)s
    %(returns)s
    """
    # Get dims
    n_features, n_fsets = adj.shape
    # Add intercept
    adj = np.column_stack((np.ones((n_features, )), adj))
    # Compute inv and df for lm
    inv = np.linalg.inv(np.dot(adj.T, adj))
    df = n_features - n_fsets - 1
    m = f'mlm - fitting {n_fsets} multivariate models of {n_features} observations with {df} degrees of freedom'
    _log(m, level='info', verbose=verbose)
    # Compute tval
    es = _tval(adj, mat.T, inv, df)[:, 1:]
    # Compute pval
    pv = 2 * (1 - sts.t.cdf(x=np.abs(es), df=df))
    return es, pv


_mlm = MethodMeta(
    name='mlm',
    desc='Multivariate Linear Model (MLM)',
    func=_func_mlm,
    stype='numerical',
    adj=True,
    weight=True,
    test=True,
    limits=(-np.inf, +np.inf),
    reference='https://doi.org/10.1093/bioadv/vbac016',
)
mlm = Method(_method=_mlm)



import warnings

import pandas as pd
import pytest
import anndata as ad

import decoupler as dc


@pytest.mark.parametrize(
    'url', [
        'https://datasets.cellxgene.cziscience.com/' +
        'f665effe-d95a-4211-ab03-9d1777ca0806.h5ad',
        'https://datasets.cellxgene.cziscience.com/' +
        '1338d08a-481a-426c-ad60-9f4ac08afe16.h5ad'
    ]
)
def test_download_anndata(
    url
):
    warnings.filterwarnings("ignore", module="anndata")
    adata = dc.ds._scell._download_anndata(url=url)
    assert isinstance(adata, ad.AnnData)


def test_pbmc3k():
    warnings.filterwarnings("ignore", module="anndata")
    adata = dc.ds.pbmc3k()
    assert isinstance(adata, ad.AnnData)
    assert adata.raw is None
    assert isinstance(adata.obs, pd.DataFrame)
    cols = {'celltype', 'leiden'}
    assert cols.issubset(adata.obs.columns)
    assert 'louvain' not in adata.obs.columns
    for col in cols:
        assert isinstance(adata.obs[col].dtype, pd.CategoricalDtype)


def test_covid5k():
    adata = dc.ds.covid5k()
    assert isinstance(adata, ad.AnnData)
    assert adata.raw is None
    assert isinstance(adata.obs, pd.DataFrame)
    cols = {'individual', 'sex', 'disease', 'celltype'}
    assert cols.issubset(adata.obs.columns)
    for col in cols:
        assert isinstance(adata.obs[col].dtype, pd.CategoricalDtype)


def test_erygast1k():
    adata = dc.ds.erygast1k()
    assert isinstance(adata, ad.AnnData)
    assert adata.raw is None
    assert isinstance(adata.obs, pd.DataFrame)
    cols = {'sample', 'stage', 'sequencing.batch', 'theiler', 'celltype'}
    assert cols.issubset(adata.obs.columns)
    for col in cols:
        assert isinstance(adata.obs[col].dtype, pd.CategoricalDtype)
    keys = {'X_pca', 'X_umap'}
    assert keys.issubset(adata.obsm.keys())
    


from decoupler.bm.pl._auc import auc
from decoupler.bm.pl._fscore import fscore
from decoupler.bm.pl._qrank import qrank
from decoupler.bm.pl._bar import bar
from decoupler.bm.pl._summary import summary



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import to_rgb, Normalize
from matplotlib.cm import ScalarMappable
from matplotlib.figure import Figure
import seaborn as sns
from anndata import AnnData

from decoupler._docs import docs
from decoupler._Plotter import Plotter
from decoupler.pp.anndata import get_obsm, bin_order
from decoupler.pp.net import prune


@docs.dedent
def order_targets(
    adata: AnnData,
    net: pd.DataFrame,
    source: str,
    order: str,
    score: str = 'score_ulm',
    label: str | None = None,
    nbins: int = 100,
    top: int = 10,
    pos_cmap: str = 'Reds',
    neg_cmap: str = 'Blues',
    color_score: str = '#88c544',
    vmin: int | float = None,
    vmax: int | float = None,
    **kwargs
) -> None | Figure:
    """
    Plot a source score, together with its targets readouts, along a continuous ordered process such as pseudotime.

    Parameters
    ----------
    %(adata)s
    %(net)s
    source
        Which source from ``net`` to show.
    %(order)s
    score
        ``adata.obsm`` key where enrichment scores are stored.
    %(label)s
    %(nbins)s
    top
        How many targets to show ranked by their standard deviation along the ordered process.
    pos_cmap
        Colormap for targets with positive weights.
    net_cmap
        Colormap for targets with negative weights.
    color_score
        Color used to plot the enrichment score.
    vmin
        Minimum value to color.
    vmax
        Minimum value to color.
    %(plot)s
    """
    # Validate
    assert isinstance(source, str), 'source must be str'
    assert isinstance(score, str), 'score must be str'
    assert isinstance(top, int) and top > 0, 'top must be int and > 0'
    # Filter net to adata
    snet = prune(features=adata.var_names, net=net, tmin=0)
    snet = snet[snet['source'] == source]
    assert snet.shape[0] > 0, f'{source} must be in net["source"]'
    # Get scoring method
    score = get_obsm(adata=adata, key=score)
    # Split by sign
    pos_names = snet[snet['weight'] > 0]['target'].values.astype('U')
    neg_names = snet[snet['weight'] < 0]['target'].values.astype('U')
    # Bin by order
    df_ftr = bin_order(
        adata=adata,
        order=order,
        names=snet['target'].to_list(),
        label=label,
        nbins=nbins,
    )
    df_scr = bin_order(
        adata=score,
        order=order,
        names=source,
        label=label,
        nbins=nbins,
    )
    # Check if labels are included
    has_cbar = False
    if np.isin(['label', 'color'], df_ftr.columns).all():
        colors = df_ftr[df_ftr['name'] == df_ftr.loc[0, 'name']]['color']
        colors = [[to_rgb(c) for c in colors]]
        has_cbar = True
    # Get mat of target values
    mat = (
        df_ftr
        .groupby(['name', 'order'], as_index=False)['value'].mean()
        .pivot(index='name', columns='order', values='value')
    )
    if vmax is None:
        vmax = mat.values.max()
    if vmin is None:
        vmin = mat.values.min()
    # Sort by magnitude
    sorted_names = mat.std(1, ddof=1).sort_values().tail(top).index
    pos_names = sorted_names.intersection(pos_names)[::-1]
    neg_names = sorted_names.intersection(neg_names)
    n_names = pos_names.size + neg_names.size
    # Instance
    kwargs = kwargs.copy()
    kwargs.setdefault('figsize', (6, np.max([sorted_names.size / 3, 3])))
    kwargs['ax'] = None
    bp = Plotter(**kwargs)
    bp.fig.delaxes(bp.ax)
    plt.close(bp.fig)
    # Plot
    bp.fig, axes = plt.subplots(
        2,
        1,
        sharex=True,
        height_ratios=[0.2, 0.8],
        figsize=bp.figsize,
        dpi=bp.dpi,
    )
    # Source score
    ax = axes[0]
    sns.lineplot(
        data=df_scr,
        x='order',
        y='value',
        ax=ax,
        color=color_score,
        lw=2
    )
    ax.set_ylabel(f'{source}\nscore')
    # Target values
    ax = axes[1]
    ax.set_xlabel('order')
    ax.grid(axis='y', visible=False)
    ax.set_ylabel('Targets')
    yticklabels = []
    # Add neg targets
    if neg_names.size > 0:
        img = ax.imshow(mat.loc[neg_names], extent=[0, 1, 0, neg_names.size], aspect='auto', cmap=neg_cmap, vmin=vmin, vmax=vmax)
        yticklabels.extend(list(neg_names)[::-1])
        cbar_mappable = ScalarMappable(cmap=neg_cmap, norm=Normalize(vmin=vmin, vmax=vmax))
        pos = ax.get_position().bounds
        cax = bp.fig.add_axes([0.97, pos[1], 0.05, (pos[3] / 2) - .025])
        cax.grid(axis='y', visible=False)
        bp.fig.colorbar(cbar_mappable, cax=cax, aspect=5, shrink=0.5, label='- target\nvalues', location='right')
    ax.axhline(y=neg_names.size, c='black', lw=1)
    # Add pos targets
    if pos_names.size > 0:
        img = ax.imshow(mat.loc[pos_names], extent=[0, 1, neg_names.size, neg_names.size + pos_names.size], aspect='auto', cmap='Reds', vmin=vmin, vmax=vmax)
        yticklabels.extend(list(pos_names)[::-1])
        cbar_mappable = ScalarMappable(cmap=pos_cmap, norm=Normalize(vmin=vmin, vmax=vmax))
        pos = ax.get_position().bounds
        #  (left, bottom, width, height)
        cax = bp.fig.add_axes([0.97, pos[1] + (pos[3] / 2) + .025, 0.05, (pos[3] / 2) - .025])
        cax.grid(axis='y', visible=False)
        bp.fig.colorbar(cbar_mappable, cax=cax, aspect=5, shrink=0.5, label='+ target\nvalues', location='right')
    # Plot labels
    ax.set_ylim(0, n_names)
    if has_cbar:
        ax.imshow(colors, aspect='auto', extent=[0, 1, n_names, 1.1 * n_names], zorder=2)
        ax.axhline(y=n_names, c='black', lw=1)
        ax.set_ylim(0, 1.1 * n_names)
    # Format plot
    ax.set_yticks(np.arange(n_names) + 0.5)
    ax.set_yticklabels(yticklabels)
    bp.fig.subplots_adjust(hspace=0)
    return bp._return()



from decoupler.pl._barplot import barplot
from decoupler.pl._dotplot import dotplot
from decoupler.pl._filter_by_expr import filter_by_expr
from decoupler.pl._filter_by_prop import filter_by_prop
from decoupler.pl._leading_edge import leading_edge
from decoupler.pl._network import network
from decoupler.pl._obsbar import obsbar
from decoupler.pl._order_targets import order_targets
from decoupler.pl._order import order
from decoupler.pl._obsm import obsm
from decoupler.pl._filter_samples import filter_samples
from decoupler.pl._source_targets import source_targets
from decoupler.pl._volcano import volcano



from typing import Tuple, Callable

import pandas as pd
import numpy as np
from anndata import AnnData
import scipy.sparse as sps
import scipy.stats as sts
from tqdm.auto import tqdm

from decoupler._log import _log
from decoupler._datatype import DataType
from decoupler.pp.net import prune, adjmat, idxmat
from decoupler.pp.data import extract


def _return(
    name: str,
    data: DataType,
    es: pd.DataFrame,
    pv: pd.DataFrame,
    verbose: bool = False,
) -> Tuple[pd.DataFrame, pd.DataFrame] | AnnData | None:
    if isinstance(data, AnnData):
        if data.obs_names.size != es.index.size:
            m = 'Provided AnnData contains empty observations, returning repaired object'
            _log(m, level='warn', verbose=verbose)
            data = data[es.index, :].copy()
            data.obsm[f'score_{name}'] = es
            if pv is not None:
                data.obsm[f'padj_{name}'] = pv
            return data
        else:
            data.obsm[f'score_{name}'] = es
            if pv is not None:
                data.obsm[f'padj_{name}'] = pv
            return None
    else:
        return es, pv


def _run(
    name: str,
    func: Callable,
    adj: bool,
    test: bool,
    data: DataType,
    net: pd.DataFrame,
    tmin: int | float = 5,
    layer: str | None = None,
    raw: bool = False,
    empty: bool = True,
    bsize: int | float = 250_000,
    verbose: bool = False,
    **kwargs
) -> Tuple[pd.DataFrame, pd.DataFrame] | AnnData | None:
    _log(f'{name} - Running {name}', level='info', verbose=verbose)
    # Process data
    mat, obs, var = extract(data, layer=layer, raw=raw, empty=empty, verbose=verbose)
    sparse = sps.issparse(mat)
    # Process net
    net = prune(features=var, net=net, tmin=tmin, verbose=verbose)
    # Handle stat type
    if adj:
        sources, targets, adjm = adjmat(features=var, net=net, verbose=verbose)
        # Handle sparse
        if sparse:
            nbatch = int(np.ceil(obs.size / bsize))
            es, pv = [], []
            for i in tqdm(range(nbatch), disable=not verbose):
                srt, end = i * bsize, i * bsize + bsize
                bmat = mat[srt:end].toarray()
                bes, bpv = func(bmat, adjm, verbose=verbose, **kwargs)
                es.append(bes)
                pv.append(bpv)
            es = np.vstack(es)
            es = pd.DataFrame(es, index=obs, columns=sources)
        else:
            es, pv = func(mat, adjm, verbose=verbose, **kwargs)
            es = pd.DataFrame(es, index=obs, columns=sources)
    else:
        sources, cnct, starts, offsets = idxmat(features=var, net=net, verbose=verbose)
        es, pv = func(mat, cnct, starts, offsets, verbose=verbose, **kwargs)
        es = pd.DataFrame(es, index=obs, columns=sources)
    # Handle pvals and FDR correction
    if test:
        pv = np.vstack(pv)
        pv = pd.DataFrame(pv, index=obs, columns=sources)
        if name != 'mlm':
            _log(f'{name} - adjusting p-values by FDR', level='info', verbose=verbose)
            pv.loc[:, :] = sts.false_discovery_control(pv.values, axis=1, method='bh')
    else:
        pv = None
    _log(f'{name} - done', level='info', verbose=verbose)
    return _return(name, data, es, pv, verbose=verbose)



from typing import Tuple

import numpy as np

from decoupler.bm._pp import _validate_bool


def _binary_clf_curve(
    y_true: np.ndarray,
    y_score: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    # Sort scores
    idx = np.flip(np.argsort(y_score))
    y_score = y_score[idx]
    y_true = y_true[idx]
    # Find unique value idxs
    idx = np.where(np.diff(y_score))[0]
    # Append a value for the end of the curve
    idx = np.append(idx, y_true.size - 1)
    # Acucmulate TP with decreasing threshold
    tps = np.cumsum(y_true)[idx]
    fps = 1 + idx - tps
    return fps, tps, y_score[idx]


def auroc(
    y_true: np.ndarray,
    y_score: np.ndarray,
) -> float:
    _validate_bool(y_true=y_true, y_score=y_score)
    # Compute binary curve
    fps, tps, thr = _binary_clf_curve(y_true, y_score)
    # Add limits
    fps = np.append(0., fps)
    tps = np.append(0., tps)
    thr = np.append(thr[0] + 1., thr)
    # Compute ratios
    fpr = fps / fps[-1]
    tpr = tps / tps[-1]
    # Compute auc
    dx = np.diff(np.ascontiguousarray(fpr))
    # Get direction slope
    if np.all(dx <= 0):
        d = -1.
    else:
        d = 1.
    # Compute area
    ret = np.sum((dx * (tpr[1:] + tpr[:-1]) / 2.0))
    auc = d * ret
    return auc


def auprc(
    y_true: np.ndarray,
    y_score: np.ndarray,
    pi0: float = 0.5
) -> float:
    _validate_bool(y_true=y_true, y_score=y_score)
    assert isinstance(pi0, (int, float)) and 0. <= pi0 <= 1., \
    'pi0 must be numeric and between 0 and 1'
    # Compute binary curve
    fps, tps, thr = _binary_clf_curve(y_true, y_score)
    # Compute prc
    ps = tps + fps
    msk = ps != 0
    if pi0 > 0 and pi0 < 1:
        # Siblini W., Fréry J., He-Guelton L., Oblé F., Wang YQ. (2020) Master
        # Your Metrics with Calibration. In: Berthold M., Feelders A., Krempl G.
        # (eds) Advances in Intelligent Data Analysis XVIII. IDA 2020. Lecture
        # Notes in Computer Science, vol 12080. Springer, Cham
        pi = np.sum(y_true) / y_true.size
        ratio = pi * (1 - pi0) / (pi0 * (1 - pi))
        prc = tps[msk] / (tps[msk] + ratio * fps[msk])
    else:
        prc = np.divide(tps[msk], ps[msk])
    # Compute rcl
    rcl = tps / tps[-1]
    # Flip and add limits
    prc = np.append(np.flip(prc), 1)
    rcl = np.append(np.flip(rcl), 0)
    thr = np.flip(thr)
    dx = np.diff(np.ascontiguousarray(rcl))
    auc = -np.sum(dx * prc[:-1])
    return auc


def auc(
    y_true: np.ndarray,
    y_score: np.ndarray,
    pi0: float = 0.5,
) -> Tuple[float, float]:
    """
    Area Under the Curve.
    """
    # Normalize to make comparable
    norm = np.nanmax(np.abs(y_score), axis=1)
    y_score = y_score / norm.reshape(-1, 1)
    assert ((-1. <= y_score) & (y_score <= 1.)).all()
    # Flatten and remove nans
    y_true, y_score = y_true.ravel(), y_score.ravel()
    msk_nan = ~np.isnan(y_score)
    y_true, y_score = y_true[msk_nan], y_score[msk_nan]
    auc_roc = auroc(y_true=y_true, y_score=y_score)
    auc_prc = auprc(y_true=y_true, y_score=y_score, pi0=pi0)
    return auc_roc, auc_prc

auc.scores = ['auroc', 'auprc']



import pandas as pd
import numpy as np

from decoupler.bm.pl._format import _format


def _hmean(
    x: float | int,
    y: float | int,
    beta: float | int = 1,
) -> float:
    assert isinstance(beta, (int, float)) and 0 < beta, \
    'beta must be numeric and > 0'
    h = np.zeros(len(x))
    msk = (x != 0.) & (y != 0.)
    h[msk] = (1 + beta**2) * (x[msk] * y[msk]) / ((x[msk] * beta**2) + y[msk])
    return h


def hmean(
    df: pd.DataFrame,
    metrics: str | list = ['auc', 'fscore', 'qrank'],
    beta: int | float = 0.5,
) -> pd.DataFrame:
    """
    Computes the harmonic mean between two metric statistics.

    Parameters
    ----------
    %(df)s
    metrics
        Metrics which to compute the harmonic mean between their own statistics.
    beta
        Controls the balance between statistics, where beta > 1 favors the first one (for example recall),
        beta < 1 the other one (for example precision), and beta = 1 gives equal weight to both.

    Returns
    -------
    Dataframe containing the harmonic mean per metric.
    """
    d_metrics = {
        'auc': {
            'name': 'H(auroc, auprc)',
            'cols': ['auprc', 'auroc'],
        },
        'fscore': {
            'name': 'F-score',
            'cols': ['precision', 'recall'],
        },
        'qrank': {
            'name': 'H(1-qrank, -log10(pval))',
            'cols': ['-log10(pval)', '1-qrank'],
        },
    }
    hdf = []
    h_cols = []
    for i, metric in enumerate(metrics):
        # Format
        cols = d_metrics[metric]['cols']
        tmp = _format(df=df, cols=cols)
        # Compute harmonic mean
        name = d_metrics[metric]['name']
        tmp[name] = _hmean(tmp[cols[0]], tmp[cols[1]], beta=beta)
        #tmp = tmp.drop(columns=cols)
        if i == 0:
            hdf.append(tmp)
        else:
            hdf.append(tmp[cols + [name]])
        h_cols.append(name)
    hdf = pd.concat(hdf, axis=1)
    # Mean qrank (final score)
    hdf['score'] = hdf[h_cols].mean(axis=1, numeric_only=True)
    hdf['score'] = (hdf['score'] - hdf['score'].min()) / (hdf['score'].max() - hdf['score'].min())
    return hdf
        


from typing import Tuple

import numpy as np
import scipy.sparse as sps
from tqdm.auto import tqdm
from xgboost import XGBRegressor

from decoupler._docs import docs
from decoupler._log import _log
from decoupler._Method import MethodMeta, Method


def _xgbr(
    x: np.ndarray,
    y: np.ndarray,
    **kwargs,
) -> np.ndarray:
    kwargs.setdefault('n_estimators', 10)
    # Init model
    reg = XGBRegressor(**kwargs)
    # Fit
    x, y = x.reshape(-1, 1), y.reshape(-1, 1)
    reg = reg.fit(x, y)
    # Get R score
    es = reg.score(x, y)
    # Clip to [0, 1]
    es = np.clip(es, 0, 1)
    return es


@docs.dedent
def _func_udt(
    mat: np.ndarray,
    adj: np.ndarray,
    verbose: bool = False,
    **kwargs,
) -> Tuple[np.ndarray, None]:
    """
    Univariate Decision Tree (UDT) :cite:`decoupler`.

    This approach uses the molecular features from one observation as the population of samples
    and it fits a gradient boosted decision trees model with a single covariate,
    which is the feature weights of a set :math:`F`.
    It uses the implementation provided by ``xgboost`` :cite:`xgboost`.

    The enrichment score :math:`ES` is then calculated as the coefficient of determination :math:`R^2`.

    %(notest)s

    %(params)s

    kwargs
        All other keyword arguments are passed to ``xgboost.XGBRegressor``.
    %(returns)s
    """
    nobs = mat.shape[0]
    nvar, nsrc = adj.shape
    m = f'udt - fitting {nsrc} univariate decision tree models (XGBoost) of {nvar} targets across {nobs} observations' 
    _log(m, level='info', verbose=verbose)
    es = np.zeros(shape=(nobs, nsrc))
    for i in tqdm(range(nobs), disable=not verbose):
        obs = mat[i]
        for j in range(adj.shape[1]):
            es[i, j] = _xgbr(x=adj[:, j], y=obs, **kwargs)
    return es, None


_udt = MethodMeta(
    name='udt',
    desc='Univariate Decision Tree (UDT)',
    func=_func_udt,
    stype='numerical',
    adj=True,
    weight=True,
    test=False,
    limits=(0, 1),
    reference='https://doi.org/10.1093/bioadv/vbac016',
)
udt = Method(_method=_udt)



import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import pytest

import decoupler as dc


@pytest.mark.parametrize(
    'use_net,name',
    [
        [False, None],
        [True, 'T1'],
        [True, 'T2'],
        [True, 'T3'],
    ]
)
def test_volcano(
    deg,
    net,
    use_net,
    name,
):
    if not use_net:
        net = None
        name = None
    fig = dc.pl.volcano(data=deg, x='stat', y='padj', net=net, name=name, return_fig=True)
    assert isinstance(fig, Figure)
    plt.close(fig)



from typing import Tuple

import numpy as np
import scipy.stats as sts
from tqdm.auto import tqdm
import numba as nb

from decoupler._docs import docs
from decoupler._log import _log
from decoupler._Method import MethodMeta, Method


@nb.njit(cache=True)
def _get_wts_posidxs(
    wts: np.ndarray,
    idxs: np.ndarray,
    pval1: np.ndarray,
    table: np.ndarray,
    penalty: int,
) -> Tuple[np.ndarray, np.ndarray]:
    pos_idxs = np.zeros(idxs.shape[0], dtype=np.int_)
    for j in nb.prange(idxs.shape[0]):
        p = pval1[j]
        if p > 0:
            x_idx, y_idx = idxs[j]
        else:
            y_idx, x_idx = idxs[j]
        pos_idxs[j] = x_idx
        x = wts[:, x_idx]
        y = wts[:, y_idx]
        x_msk, y_msk = x != 0, y != 0
        msk = x_msk * y_msk
        x[msk] = x[msk] / (1 + np.abs(p))**(penalty/table[x_idx])
        wts[:, x_idx] = x
    return wts, pos_idxs


@nb.njit(cache=True)
def _get_tmp_idxs(
    pval: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:
    size = int(np.sum(~np.isnan(pval)) / 2)
    tmp = np.zeros((size, 2))
    idxs = np.zeros((size, 2), dtype=np.int_)
    k = 0
    for i in nb.prange(pval.shape[0]):
        for j in range(pval.shape[1]):
            if i <= j:
                x = pval[i, j]
                if not np.isnan(x):
                    y = pval[j, i]
                    if not np.isnan(y):
                        tmp[k, 0] = x
                        tmp[k, 1] = y
                        idxs[k, 0] = i
                        idxs[k, 1] = j
                        k += 1
    return tmp, idxs


@nb.njit(cache=True)
def _fill_pval_mat(
    j: int,
    reg: np.ndarray,
    n_targets: int,
    s1: np.ndarray,
    s2: np.ndarray,
) -> np.ndarray:
    n_fsets = reg.shape[1]
    col = np.full(n_fsets, np.nan)
    for k in nb.prange(n_fsets):
        if k != j:
            k_msk = reg[:, k] != 0
            nhits = k_msk.sum()
            if nhits > n_targets:
                sum1 = np.sum(reg[:, k] * s2)
                ss = np.sign(sum1)
                if ss == 0:
                    ss = 1
                sum2 = np.sum((1- np.abs(reg[k_msk, k])) * s1[k_msk])
                ww = np.ones(nhits)
                col[k] = (np.abs(sum1) + sum2 * (sum2 > 0)) / ww.size * ss * np.sqrt(ww.size)
    return col


def _get_inter_pvals(
    nes_i: np.ndarray,
    ss_i: np.ndarray,
    sub_net: np.ndarray,
    n_targets: int,
) -> np.ndarray:
    pval = np.full((sub_net.shape[1], sub_net.shape[1]), np.nan)
    for j in range(sub_net.shape[1]):
        trgt_msk = sub_net[:, j] != 0
        reg = ((sub_net[trgt_msk] != 0) * sub_net[trgt_msk, j].reshape(-1, 1))
        s2 = ss_i[trgt_msk]
        s2 = sts.rankdata(s2, method='average') / (s2.shape[0]+1) * 2 - 1
        s1 = np.abs(s2) * 2 - 1
        s1 = s1 + (1 - np.max(s1)) / 2
        s1 = sts.norm.ppf(s1/2 + 0.5)
        tmp = np.sign(nes_i[j])
        if tmp == 0:
            tmp = 1
        s2 = (sts.norm.ppf(s2/2 + 0.5) * tmp)
        pval[j] = _fill_pval_mat(j=j, reg=reg, n_targets=n_targets, s1=s1, s2=s2)
    pval = 1 - sts.norm.cdf(pval)
    return pval


def _shadow_regulon(
    nes_i: np.ndarray,
    ss_i: np.ndarray,
    net: np.ndarray,
    reg_sign: float = 1.96,
    n_targets: int | float = 10,
    penalty: int | float = 20
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    # Find significant activities
    msk_sign = np.abs(nes_i) > reg_sign
    # Filter by significance
    nes_i = nes_i[msk_sign]
    sub_net = net[:, msk_sign]
    # Init likelihood mat
    wts = np.zeros(sub_net.shape)
    wts[sub_net != 0] = 1.0
    if wts.shape[1] < 2:
        return None
    # Get significant interatcions between regulators
    pval = _get_inter_pvals(nes_i, ss_i, sub_net, n_targets=n_targets)
    # Get pairs of regulators
    tmp, idxs = _get_tmp_idxs(pval)
    if tmp.size == 0:
        return None
    pval1 = np.log10(tmp[:, 1]) - np.log10(tmp[:, 0])
    unique, counts = np.unique(idxs.flatten(), return_counts=True)
    table = np.zeros(int(unique.max()) + 1, dtype=np.int_)
    table[unique.astype(np.int_)] = counts
    # Modify interactions based on sign of pval1
    wts, pos_idxs = _get_wts_posidxs(wts, idxs, pval1, table, penalty)
    # Select only regulators with positive pval1
    pos_idxs = np.unique(pos_idxs)
    sub_net = sub_net[:, pos_idxs]
    wts = wts[:, pos_idxs]
    idxs = np.where(msk_sign)[0][pos_idxs]
    return sub_net, wts, idxs


def _aREA(
    mat: np.ndarray,
    net: np.ndarray,
    wts: None = None
) -> np.ndarray:
    if wts is None:
        wts = np.zeros(net.shape)
        wts[net != 0] = 1
    wts = wts / np.max(wts, axis=0)
    nes = np.sqrt(np.sum(wts**2, axis=0))
    wts = (wts / np.sum(wts, axis=0))
    mat = sts.rankdata(mat, method='average', axis=1) / (mat.shape[1] + 1)
    t1 = np.abs(mat - 0.5) * 2
    t1 = t1 + (1 - np.max(t1))/2
    msk = np.sum(net != 0, axis=1) >= 1
    t1, mat = t1[:, msk], mat[:, msk]
    net, wts = net[msk], wts[msk]
    t1 = sts.norm.ppf(t1)
    mat = sts.norm.ppf(mat)
    sum1 = mat.dot(wts*net)
    sum2 = t1.dot((1-np.abs(net)) * wts)
    tmp = (np.abs(sum1) + sum2 * (sum2 > 0)) * np.sign(sum1)
    nes = tmp * nes
    return nes


@docs.dedent
def _func_viper(
    mat: np.ndarray,
    adj: np.ndarray,
    pleiotropy: bool = True,
    reg_sign: float = 0.05,
    n_targets: int | float = 10,
    penalty: int | float = 20,
    verbose: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Virtual Inference of Protein-activity by Enriched Regulon analysis (VIPER) :cite:`viper`.

    %(yestest)s

    %(params)s
    
    pleiotropy
        Whether correction for pleiotropic regulation should be performed.
    reg_sign
        If ``pleiotropy``, p-value threshold for considering significant regulators.
    n_targets
        If ``pleiotropy``, integer indicating the minimal number of overlaping targets to consider for analysis.
    penalty
        If ``pleiotropy``, number higher than 1 indicating the penalty for the pleiotropic interactions. 1 = no penalty.

    %(returns)s
    """
    # Get number of batches
    n_samples = mat.shape[0]
    n_features, n_fsets = adj.shape
    m = f'viper - calculating {n_fsets} scores across {n_samples} observations'
    _log(m, level='info', verbose=verbose)
    # Compute score
    nes = _aREA(mat, adj)
    if pleiotropy:
        m = f'viper - refining scores based on pleiotropy'
        _log(m, level='info', verbose=verbose)
        reg_sign = sts.norm.ppf(1-(reg_sign / 2))
        for i in tqdm(range(nes.shape[0]), disable=not verbose):
            # Extract per sample
            ss_i = mat[i]
            nes_i = nes[i]
            # Shadow regulons
            shadow = _shadow_regulon(nes_i, ss_i, adj, reg_sign=reg_sign, n_targets=n_targets, penalty=penalty)
            if shadow is None:
                continue
            else:
                sub_net, wts, idxs = shadow
            # Recompute activity with shadow regulons and update nes
            tmp = _aREA(ss_i.reshape(1, -1), sub_net, wts=wts)[0]
            nes[i, idxs] = tmp
    # Get pvalues
    pvals = 2 * sts.norm.sf(np.abs(nes))
    return nes, pvals


_viper = MethodMeta(
    name='viper',
    desc='Virtual Inference of Protein-activity by Enriched Regulon analysis (VIPER)',
    func=_func_viper,
    stype='numerical',
    adj=True,
    weight=True,
    test=True,
    limits=(-np.inf, +np.inf),
    reference='https://doi.org/10.1038/ng.3593',
)
viper = Method(_method=_viper)



import numpy as np
import scipy.sparse as sps
import pytest

import decoupler as dc


def test_extract(
    adata,
):
    data = [adata.X, adata.obs_names, adata.var_names]
    X, obs, var = dc.pp.extract(data=data)
    assert X.shape[0] == obs.size
    assert X.shape[1] == var.size
    X, obs, var = dc.pp.extract(data=adata.to_df())
    assert X.shape[0] == obs.size
    assert X.shape[1] == var.size
    X, obs, var = dc.pp.extract(data=adata)
    assert X.shape[0] == obs.size
    assert X.shape[1] == var.size
    adata.layers['counts'] = adata.X.round()
    X, obs, var = dc.pp.extract(data=adata, layer='counts')
    assert float(np.sum(X)).is_integer()
    nadata = adata.copy()
    nadata.X = nadata.X * -1
    adata.raw = nadata
    X, obs, var = dc.pp.extract(data=adata, raw=True)
    assert (X < 0).all()



import pandas as pd

from decoupler._docs import docs
from decoupler._datatype import DataType
from decoupler.mt._methods import _methods
from decoupler.mt._consensus import consensus


@docs.dedent
def decouple(
    data: DataType,
    net: pd.DataFrame,
    methods: str | list = 'all',
    args: dict = dict(),
    cons: bool = False,
    **kwargs
) -> dict | None:
    """
    Runs multiple enrichment methods sequentially.

    Parameters
    ----------
    %(data)s
    %(net)s
    methods
        List of methods to run.
    args
        Dictionary of dictionaries containing method-specific keyword arguments.
    cons
        Whether to get a consensus score across the used methods.
    %(tmin)s
    %(raw)s
    %(empty)s
    %(bsize)s
    %(verbose)s
    """
    # Validate
    _mdict = {m.name: m for m in _methods}
    if isinstance(methods, str):
        if methods == 'all':
            methods = _mdict.keys()
        else:
            methods = [methods]
    methods = set(methods)
    assert methods.issubset(_mdict), \
    f'methods={methods} must be in decoupler.\nUse decoupler.mt.show_methods to check which ones are available'
    assert all(k in methods for k in args), \
    f'All keys in args={args.keys()} must belong to a method in methods={methods}'
    kwargs = kwargs.copy()
    kwargs.setdefault('verbose', False)
    # Run each method
    all_res = {}
    for name in methods:
        mth = _mdict[name]
        arg = args.setdefault(name, {})
        res = mth(data=data, net=net, **arg, **kwargs)
        if res:
            res = {
                f'score_{mth.name}': res[0],
                f'padj_{mth.name}': res[1],
            }
            all_res = all_res | res
    if all_res:
        if cons:
            all_res['score_consensus'], all_res['padj_consensus'] = consensus(all_res, verbose=kwargs['verbose'])
        return all_res
    elif cons:
        consensus(data, verbose=kwargs['verbose'])
